//--------------------------------------
// Parametric GA for Self-Assembly System
// Atsushi Masumori 2012,
//--------------------------------------

import java.util.Comparator;
import java.util.Random;
import controlP5.*;

ArrayList<Indivisual> indivisuals = new ArrayList<Indivisual>();
ArrayList<Indivisual> nextGenetation = new ArrayList<Indivisual>();
Indivisual winner;
Indivisual elite;

ControlP5 controlP5;

float crossoverRate = 0.3;
float mutationRate = 0.05;
int numberOfTournaments = 10;
int numberOfGenes = 22;
int numberOfIndivisuals = 300;
int col = 3;
int row = 3;

int eliteID = 0;
boolean isDraw = true;
int countOfGenerarion = 0;

int seed = 200;
Random rnd = new Random(seed);

public void setup() {
  size(1500, 300, P2D);
  
  controlP5 = new ControlP5(this);  
  controlP5.addSlider("crossoverRate", 0, 1, 10, 20, 200, 15);
  controlP5.addSlider("mutationRate", 0,0.1,10,40,200,15);
  controlP5.addSlider("numberOfTournaments", 0,100,10,60,200,15);
  controlP5.addSlider("numberOfIndivisuals", 0,1000,10,80,200,15);
  
  for (int i=0; i<numberOfIndivisuals; i++) {
    indivisuals.add(new Indivisual(numberOfGenes, col, row));
    indivisuals.get(i).evaluateFitness();
  }
}

void draw() {
  if (isDraw == true) background(0);
  evaluation();
  selection();
  output();
  crossover();
  mutation();
}


void evaluation() {
  for (int i=0; i<indivisuals.size(); i++) {
    indivisuals.get(i).evaluateFitness();
  }
}


void selection() {
  nextGenetation = new ArrayList<Indivisual>();
  double currentFitness = -10000;

  //Tournament selection
  winner = new Indivisual(numberOfGenes, col, row);
  winner.fitness = -10000;
  Indivisual current = new Indivisual(numberOfGenes, col, row);

  for (int i=0; i<indivisuals.size(); i++) {
    for (int j=0; j<numberOfTournaments; j++) {
      current.set(indivisuals.get(rnd.nextInt(indivisuals.size())));
      if(current.fitness > winner.fitness) winner.set(current);
      else if (winner.fitness == current.fitness && current.fitness2 > winner.fitness2) winner.set(current);
    } 
    nextGenetation.add(winner);
  }

  for (int i=0; i<indivisuals.size(); i++) {
    indivisuals.get(i).set(nextGenetation.get(i));
  }
  
  getElite();
}

void getElite() {
  
  ArrayList<Integer> tmpEliteIDs = new ArrayList<Integer>();
  double maxFitness = -1000;
  double maxFitness2 = -1000;

  for (int i=0; i<indivisuals.size(); i++) {
    if (maxFitness <= indivisuals.get(i).fitness) {
      maxFitness = indivisuals.get(i).fitness;
      tmpEliteIDs.add(i);
    }
  }

  for (int i=0; i<tmpEliteIDs.size(); i++) {
    if (maxFitness == indivisuals.get(tmpEliteIDs.get(i)).fitness && maxFitness2 < indivisuals.get(tmpEliteIDs.get(i)).fitness2) {
      eliteID = i;
      maxFitness2 = indivisuals.get(tmpEliteIDs.get(i)).fitness2;
    }
  }
  
  elite = new Indivisual(indivisuals.get(eliteID));
}

void output() {

  if (isDraw == true) {
    for (int i=0; i<numberOfGenes/2; i++) {
      for (int j=0; j<col; j++) {
        for (int k=0; k<row; k++) { 
          //          print(indivisuals.get(numm).pattern[i][j][k]);
          fill(255);
          rect(j*20 + i*100 + 20, k*20+120, 20, 20);
          if (elite.pattern[i][j][k]==1) fill(150);
          else if (elite.pattern[i][j][k]==2) fill(0);
          if (elite.pattern[i][j][k] !=0 ) rect(j*20 + i*100 + 22, k*20+122, 16, 16);
        }
        //        println();
      }
      //      println("--");
    }
  }

  println("fitness  of elite : " + elite.fitness);
  println("fitness2 of elite : " + elite.fitness2);
  println("Generarion        : " + countOfGenerarion++);
  
  if (elite.fitness == 0 && elite.fitness2 == 0) noLoop();
}


void crossover() {
  int random1;
  int random2;
  int random3;
  Indivisual child1;
  Indivisual child2;
  ArrayList<Indivisual> children = new ArrayList<Indivisual>();

  Indivisual elite1 = new Indivisual(elite);
  Indivisual elite2 = new Indivisual(elite);
  children.add(elite1);
  children.add(elite2);
  int numberOfSpace = (indivisuals.size()-1)/2;

//  int numberOfSpace = (indivisuals.size())/2; 
  
  while (numberOfSpace>0) {    
    random1 = rnd.nextInt(indivisuals.size());
    random2 = rnd.nextInt(indivisuals.size());
    child1 = new Indivisual(indivisuals.get(random1));
    child2 = new Indivisual(indivisuals.get(random2));  

    if (rnd.nextDouble() < crossoverRate) { 
      for (int i=0; i<numberOfGenes/2; i++) {
        random3 = rnd.nextInt(numberOfGenes/2);
        for (int j=0; j<col; j++) {
          if (rnd.nextInt(10)%2 == 0) {  
            for (int k=0; k<row; k++) { 
              child1.pattern[i][j][k] = indivisuals.get(random2).pattern[random3][j][k]; 
              child2.pattern[random3][j][k] = indivisuals.get(random1).pattern[i][j][k];
            }
          }
        }
      }
    }
    children.add(child1);
    children.add(child2);
    numberOfSpace = numberOfSpace - 1;
  }
  for (int i=0; i<numberOfIndivisuals; i++) {
    if(i >= indivisuals.size() || i >= children.size()) indivisuals.add(new Indivisual(numberOfGenes, col, row));
    else indivisuals.get(i).set(children.get(i));
  }
}

void mutation() {  
  int[][][]tmp = new int[numberOfGenes/2][col][row];
  
  for (int i=0; i<indivisuals.size(); i++) {
    if (i != eliteID) {
      for (int l=0; l<numberOfGenes/2; l++) {
        for (int m=0; m<col; m++) {
          for (int n=0; n<row; n++) {
            if ((m != 1 || n != 1 ) && rnd.nextDouble() < mutationRate) indivisuals.get(i).pattern[l][m][n] = rnd.nextInt(3);
          }
        }
      }
    }
  }
}

void keyPressed() {
  if (key == 'd') isDraw = !isDraw;
}

